import pandas as pd
from pandas.testing import assert_frame_equal
from pytest import fixture, raises

from otoole.results.result_package import (
    ResultsPackage,
    capital_recovery_factor,
    discount_factor,
    discount_factor_storage,
    discount_factor_storage_salvage,
    pv_annuity,
)


@fixture
def new_capacity():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.3],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1.6],
            ["SIMPLICITY", "DUMMY", 2014, 0.9],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return df


@fixture
def new_capacity_bitty():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 0.0300000000000001],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 0.0309999999999999],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 0.032],
            ["SIMPLICITY", "GAS_EXTRACTION", 2017, 0.0329999999999999],
            ["SIMPLICITY", "GAS_EXTRACTION", 2018, 0.0330000000000002],
            ["SIMPLICITY", "DUMMY", 2014, 0.9],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return df


@fixture
def operational_life():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", 2], ["SIMPLICITY", "DUMMY", 3]],
        columns=["REGION", "TECHNOLOGY", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY"])
    return df


@fixture
def operational_life_overlap():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", 30], ["SIMPLICITY", "DUMMY", 30]],
        columns=["REGION", "TECHNOLOGY", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY"])
    return df


@fixture
def rate_of_activity():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "ID", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "IN", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "SD", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "SN", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "WD", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "WN", "GAS_EXTRACTION", 1, 2014, 1],
        ],
        columns=[
            "REGION",
            "TIMESLICE",
            "TECHNOLOGY",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    )
    return df.set_index(
        ["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"]
    )


@fixture
def rate_of_activity_two_tech():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "ID", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "IN", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "SD", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "SN", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "WD", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "WN", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "ID", "DUMMY", 1, 2014, 1],
            ["SIMPLICITY", "IN", "DUMMY", 1, 2014, 1],
            ["SIMPLICITY", "SD", "DUMMY", 1, 2014, 1],
            ["SIMPLICITY", "SN", "DUMMY", 1, 2014, 1],
            ["SIMPLICITY", "WD", "DUMMY", 1, 2014, 1],
            ["SIMPLICITY", "WN", "DUMMY", 1, 2014, 1],
        ],
        columns=[
            "REGION",
            "TIMESLICE",
            "TECHNOLOGY",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    )
    return df.set_index(
        ["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"]
    )


@fixture
def year():
    return pd.DataFrame(
        data=[2014, 2015, 2016, 2017, 2018, 2019, 2020], columns=["VALUE"]
    )


@fixture
def region():
    return pd.DataFrame(data=["SIMPLICITY"], columns=["VALUE"])


@fixture
def region_multiple():
    return pd.DataFrame(data=["SIMPLICITY", "DUMMY"], columns=["VALUE"])


@fixture
def storage():
    return pd.DataFrame(data=["DAM"], columns=["VALUE"])


@fixture
def accumulated_new_capacity():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.3],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 1.3],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1.6],
            ["SIMPLICITY", "GAS_EXTRACTION", 2017, 1.6],
            ["SIMPLICITY", "DUMMY", 2014, 0.9],
            ["SIMPLICITY", "DUMMY", 2015, 0.9],
            ["SIMPLICITY", "DUMMY", 2016, 0.9],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def residual_capacity():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 0],
            ["SIMPLICITY", "GAS_EXTRACTION", 2017, 0],
            ["SIMPLICITY", "DUMMY", 2014, 0.1],
            ["SIMPLICITY", "DUMMY", 2015, 0.2],
            ["SIMPLICITY", "DUMMY", 2016, 0.3],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def total_capacity():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 2.3],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 2.3],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1.6],
            ["SIMPLICITY", "GAS_EXTRACTION", 2017, 1.6],
            ["SIMPLICITY", "DUMMY", 2014, 1.0],
            ["SIMPLICITY", "DUMMY", 2015, 1.1],
            ["SIMPLICITY", "DUMMY", 2016, 1.2],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def capital_cost():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.23],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 2.34],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 3.45],
            ["SIMPLICITY", "DUMMY", 2014, 4.56],
            ["SIMPLICITY", "DUMMY", 2015, 5.67],
            ["SIMPLICITY", "DUMMY", 2016, 6.78],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return df


@fixture
def fixed_cost():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 2017, 1],
            ["SIMPLICITY", "DUMMY", 2014, 0.5],
            ["SIMPLICITY", "DUMMY", 2015, 0.5],
            ["SIMPLICITY", "DUMMY", 2016, 0.5],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def variable_cost():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", 1, 2014, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 1, 2015, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 1, 2016, 1],
            ["SIMPLICITY", "GAS_EXTRACTION", 1, 2017, 1],
            ["SIMPLICITY", "DUMMY", 1, 2014, 0.5],
            ["SIMPLICITY", "DUMMY", 1, 2015, 0.5],
            ["SIMPLICITY", "DUMMY", 1, 2016, 0.5],
        ],
        columns=["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"])
    return data


@fixture
def undiscounted_capital_investment():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 10],
            ["SIMPLICITY", "DUMMY", 2015, 0],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 123],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 456],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 789],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def annual_fixed_operating_cost():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 10],
            ["SIMPLICITY", "DUMMY", 2015, 0],
            ["SIMPLICITY", "DUMMY", 2016, 10],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 123],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 456],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 789],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def annual_variable_operating_cost():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 10],
            ["SIMPLICITY", "DUMMY", 2015, 10],
            ["SIMPLICITY", "DUMMY", 2016, 0],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 321],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 654],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 987],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def discounted_capital_costs():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 10],
            ["SIMPLICITY", "DUMMY", 2015, 0],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 111],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 222],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 333],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def discounted_operational_costs():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 5],
            ["SIMPLICITY", "DUMMY", 2015, 10],
            ["SIMPLICITY", "DUMMY", 2016, 20],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 444],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 555],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 666],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def discounted_emissions_penalty():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 10],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 777],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def discounted_salvage_value():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 1],
            ["SIMPLICITY", "DUMMY", 2015, 2],
            ["SIMPLICITY", "DUMMY", 2016, 3],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 888],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 999],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def discounted_technology_cost():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DUMMY", 2014, 111],
            ["SIMPLICITY", "DUMMY", 2015, 222],
            ["SIMPLICITY", "DUMMY", 2016, 333],
            ["SIMPLICITY", "GAS_EXTRACTION", 2014, 444],
            ["SIMPLICITY", "GAS_EXTRACTION", 2015, 555],
            ["SIMPLICITY", "GAS_EXTRACTION", 2016, 666],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    return data


@fixture
def capital_cost_storage():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DAM", 2014, 1.23],
            ["SIMPLICITY", "DAM", 2015, 2.34],
            ["SIMPLICITY", "DAM", 2016, 3.45],
            ["SIMPLICITY", "BATTERY", 2014, 4.56],
            ["SIMPLICITY", "BATTERY", 2015, 5.67],
            ["SIMPLICITY", "BATTERY", 2016, 6.78],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return df


@fixture
def new_storage_capacity():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DAM", 2014, 1.3],
            ["SIMPLICITY", "DAM", 2016, 1.6],
            ["SIMPLICITY", "BATTERY", 2014, 0.9],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return df


@fixture
def undiscounted_capital_investment_storage():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DAM", 2014, 1.23],
            ["SIMPLICITY", "DAM", 2015, 2.34],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return data


@fixture
def salvage_value_storage():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DAM", 2014, 1.23],
            ["SIMPLICITY", "DAM", 2015, 2.34],
            ["SIMPLICITY", "DAM", 2016, 3.45],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return data


@fixture
def discounted_capital_costs_storage():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "BATTERY", 2014, 11.1],
            ["SIMPLICITY", "BATTERY", 2016, 22.2],
            ["SIMPLICITY", "DAM", 2014, 33.3],
            ["SIMPLICITY", "DAM", 2015, 44.4],
            ["SIMPLICITY", "DAM", 2016, 55.5],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return data


@fixture
def discounted_salvage_value_storage():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DAM", 2014, 1.23],
            ["SIMPLICITY", "DAM", 2015, 2.34],
            ["SIMPLICITY", "DAM", 2016, 3.45],
            ["SIMPLICITY", "BATTERY", 2014, 4.56],
            ["SIMPLICITY", "BATTERY", 2015, 5.67],
            ["SIMPLICITY", "BATTERY", 2016, 6.78],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return data


@fixture
def discounted_storage_cost():
    data = pd.DataFrame(
        data=[
            ["SIMPLICITY", "DAM", 2014, 11.1],
            ["SIMPLICITY", "DAM", 2015, 22.2],
            ["SIMPLICITY", "DAM", 2016, 33.3],
            ["SIMPLICITY", "BATTERY", 2014, 44.4],
            ["SIMPLICITY", "BATTERY", 2015, 55.5],
            ["SIMPLICITY", "BATTERY", 2016, 66.6],
        ],
        columns=["REGION", "STORAGE", "YEAR", "VALUE"],
    ).set_index(["REGION", "STORAGE", "YEAR"])
    return data


@fixture(scope="function")
def null() -> ResultsPackage:
    package = ResultsPackage({})
    return package


@fixture(scope="function")
def minimal(emission_activity_ratio, yearsplit, rate_of_activity) -> ResultsPackage:

    dummy_results = {
        "EmissionActivityRatio": emission_activity_ratio,
        "YearSplit": yearsplit,
        "RateOfActivity": rate_of_activity,
    }
    package = ResultsPackage(dummy_results)
    return package


@fixture(scope="function")
def two_tech(
    emission_activity_ratio, yearsplit, rate_of_activity_two_tech
) -> ResultsPackage:
    dummy_results = {
        "EmissionActivityRatio": emission_activity_ratio,
        "YearSplit": yearsplit,
        "RateOfActivity": rate_of_activity_two_tech,
    }
    package = ResultsPackage(dummy_results)
    return package


@fixture(scope="function")
def two_tech_emission_activity(
    emission_activity_ratio_two_techs, yearsplit, rate_of_activity_two_tech
) -> ResultsPackage:
    dummy_results = {
        "EmissionActivityRatio": emission_activity_ratio_two_techs,
        "YearSplit": yearsplit,
        "RateOfActivity": rate_of_activity_two_tech,
    }
    package = ResultsPackage(dummy_results)
    return package


class TestCalculateAnnualEmissions:
    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.annual_emissions()
        assert "Cannot calculate AnnualEmissions due to missing data" in str(ex.value)

    def test_minimal(self, minimal: ResultsPackage):
        """ """
        package = minimal
        actual = package.annual_emissions()

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "CO2", 2014, 1.0]],
            columns=["REGION", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "EMISSION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_missing_tech(self, two_tech):
        """ """
        package = two_tech
        actual = package.annual_emissions()

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "CO2", 2014, 1.0]],
            columns=["REGION", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "EMISSION", "YEAR"])
        assert_frame_equal(actual, expected)


class TestCalculateAnnualTechnologyEmissions:
    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.annual_technology_emissions()
        assert "Cannot calculate AnnualTechnologyEmission due to missing data" in str(
            ex
        )

    def test_minimal(self, two_tech):
        """ """
        package = two_tech
        actual = package.annual_technology_emissions()

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 2014, 1.0]],
            columns=["REGION", "TECHNOLOGY", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_no_zeros(self, two_tech_emission_activity):
        """ """
        package = two_tech_emission_activity
        actual = package.annual_technology_emissions()

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 2014, 1.0]],
            columns=["REGION", "TECHNOLOGY", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "YEAR"])

        assert_frame_equal(actual, expected)


class TestCalculateAnnualTechnologyEmissionsByMode:
    def test_null(self, null):
        """ """
        package = null

        with raises(KeyError) as ex:
            package.annual_technology_emission_by_mode()
        assert (
            "Cannot calculate AnnualTechnologyEmissionByMode due to missing data"
            in str(ex.value)
        )

    def test_minimal(self, two_tech):
        """ """
        package = two_tech
        actual = package.annual_technology_emission_by_mode()
        expected = pd.DataFrame(
            data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
            columns=[
                "REGION",
                "TECHNOLOGY",
                "EMISSION",
                "MODE_OF_OPERATION",
                "YEAR",
                "VALUE",
            ],
        ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_no_zeros(self, two_tech_emission_activity):
        """ """
        package = two_tech_emission_activity
        actual = package.annual_technology_emission_by_mode()

        expected = pd.DataFrame(
            data=[
                # ['SIMPLICITY', 'DUMMY', 'CO2', 2014, 0.0],
                ["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]
            ],
            columns=[
                "REGION",
                "TECHNOLOGY",
                "EMISSION",
                "MODE_OF_OPERATION",
                "YEAR",
                "VALUE",
            ],
        ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])

        assert_frame_equal(actual, expected)


class TestDiscountedTechnologyEmissionsPenalty:
    def test_calculate(
        self,
        emissions_penalty,
        region,
        year,
        discount_rate,
        annual_technology_emissions_by_mode,
    ):

        results = {
            "AnnualTechnologyEmissionByMode": annual_technology_emissions_by_mode,
            "EmissionsPenalty": emissions_penalty,
            "REGION": region,
            "YEAR": year,
            "DiscountRate": discount_rate,
        }
        package = ResultsPackage(results)
        actual = package.discounted_tech_emis_pen()

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.2003570897266957]],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)


class TestAccumulatedNewCapacity:
    def test_individual(self, new_capacity, operational_life, year):

        results = {
            "NewCapacity": new_capacity,
            "OperationalLife": operational_life,
            "YEAR": year,
        }
        package = ResultsPackage(results)

        actual = package.accumulated_new_capacity()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 1.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1.6],
                ["SIMPLICITY", "GAS_EXTRACTION", 2017, 1.6],
                ["SIMPLICITY", "DUMMY", 2014, 0.9],
                ["SIMPLICITY", "DUMMY", 2015, 0.9],
                ["SIMPLICITY", "DUMMY", 2016, 0.9],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_overlapping(self, new_capacity, operational_life_overlap, year):

        results = {
            "NewCapacity": new_capacity,
            "OperationalLife": operational_life_overlap,
            "YEAR": year,
        }
        package = ResultsPackage(results)

        actual = package.accumulated_new_capacity()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 1.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 2.9],
                ["SIMPLICITY", "GAS_EXTRACTION", 2017, 2.9],
                ["SIMPLICITY", "GAS_EXTRACTION", 2018, 2.9],
                ["SIMPLICITY", "GAS_EXTRACTION", 2019, 2.9],
                ["SIMPLICITY", "GAS_EXTRACTION", 2020, 2.9],
                ["SIMPLICITY", "DUMMY", 2014, 0.9],
                ["SIMPLICITY", "DUMMY", 2015, 0.9],
                ["SIMPLICITY", "DUMMY", 2016, 0.9],
                ["SIMPLICITY", "DUMMY", 2017, 0.9],
                ["SIMPLICITY", "DUMMY", 2018, 0.9],
                ["SIMPLICITY", "DUMMY", 2019, 0.9],
                ["SIMPLICITY", "DUMMY", 2020, 0.9],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_overlapping_bitty(
        self, new_capacity_bitty, operational_life_overlap, year
    ):
        results = {
            "NewCapacity": new_capacity_bitty,
            "OperationalLife": operational_life_overlap,
            "YEAR": year,
        }
        package = ResultsPackage(results)

        actual = package.accumulated_new_capacity()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 0.0300000000000001],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 0.0609999999999999],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 0.093],
                ["SIMPLICITY", "GAS_EXTRACTION", 2017, 0.126],
                ["SIMPLICITY", "GAS_EXTRACTION", 2018, 0.159],
                ["SIMPLICITY", "GAS_EXTRACTION", 2019, 0.159],
                ["SIMPLICITY", "GAS_EXTRACTION", 2020, 0.159],
                ["SIMPLICITY", "DUMMY", 2014, 0.9],
                ["SIMPLICITY", "DUMMY", 2015, 0.9],
                ["SIMPLICITY", "DUMMY", 2016, 0.9],
                ["SIMPLICITY", "DUMMY", 2017, 0.9],
                ["SIMPLICITY", "DUMMY", 2018, 0.9],
                ["SIMPLICITY", "DUMMY", 2019, 0.9],
                ["SIMPLICITY", "DUMMY", 2020, 0.9],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)


class TestComputeFixedandVariableOperatingCost:
    """

    Notes
    -----
    1.3 + 1.0 * 1.0 = 2.3
    1.3 + 1.0 * 1.0 = 2.3
    1.6 + 0.0 * 1.0 = 1.6
    1.6 + 0.0 * 1.0 = 1.6
    0.9 + 0.1 * 0.5 = 0.5
    0.9 + 0.2 * 0.5 = 0.55
    0.9 + 0.3 * 0.5 = 0.6
    """

    def test_fixed(self, total_capacity, fixed_cost):

        results = {"TotalCapacityAnnual": total_capacity, "FixedCost": fixed_cost}
        package = ResultsPackage(results)
        actual = package.annual_fixed_operating_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 2.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 2.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1.6],
                ["SIMPLICITY", "GAS_EXTRACTION", 2017, 1.6],
                ["SIMPLICITY", "DUMMY", 2014, 0.5],
                ["SIMPLICITY", "DUMMY", 2015, 0.55],
                ["SIMPLICITY", "DUMMY", 2016, 0.6],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null_fixed(self, null):
        package = null
        with raises(KeyError) as ex:
            package.annual_fixed_operating_cost()
        assert "Cannot calculate AnnualFixedOperatingCost due to missing data" in str(
            ex.value
        )

    def test_variable(self, rate_of_activity_two_tech, yearsplit, variable_cost):
        """

        ["ID", 2014, 0.1667],
        ["IN", 2014, 0.0833],
        ["SD", 2014, 0.1667],
        ["SN", 2014, 0.0833],
        ["WD", 2014, 0.3333],
        ["WN", 2014, 0.1667],

        """
        results = {
            "RateOfActivity": rate_of_activity_two_tech,
            "YearSplit": yearsplit,
            "VariableCost": variable_cost,
        }
        package = ResultsPackage(results)
        actual = package.annual_variable_operating_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 0.5],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.0],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        assert_frame_equal(actual, expected)

    def test_null_variable(self, null):
        package = null
        with raises(KeyError) as ex:
            package.annual_variable_operating_cost()
        assert (
            "Cannot calculate AnnualVariableOperatingCost due to missing data"
            in str(ex.value)
        )


class TestComputeTotalAnnualCapacity:
    """

    Notes
    -----
    1.3 + 1.0
    1.3 + 1.0
    1.6 + 0.0
    1.6 + 0.0
    0.9 + 0.1
    0.9 + 0.2
    0.9 + 0.3
    """

    def test_compute(self, accumulated_new_capacity, residual_capacity):

        results = {
            "AccumulatedNewCapacity": accumulated_new_capacity,
            "ResidualCapacity": residual_capacity,
        }
        package = ResultsPackage(results)

        actual = package.total_capacity_annual()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 2.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 2.3],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1.6],
                ["SIMPLICITY", "GAS_EXTRACTION", 2017, 1.6],
                ["SIMPLICITY", "DUMMY", 2014, 1.0],
                ["SIMPLICITY", "DUMMY", 2015, 1.1],
                ["SIMPLICITY", "DUMMY", 2016, 1.2],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null):
        package = null
        with raises(KeyError) as ex:
            package.total_capacity_annual()
        assert "Cannot calculate TotalCapacityAnnual due to missing data" in str(
            ex.value
        )


class TestCapitalInvestment:
    def test_calculate_captital_investment_with_dr_idv(
        self,
        capital_cost,
        new_capacity,
        operational_life,
        region,
        year,
        discount_rate,
        discount_rate_idv,
    ):

        results = {
            "CapitalCost": capital_cost,
            "NewCapacity": new_capacity,
            "OperationalLife": operational_life,
            "REGION": region,
            "YEAR": year,
            "DiscountRate": discount_rate,
            "DiscountRateIdv": discount_rate_idv,
        }

        package = ResultsPackage(results)
        actual = package.capital_investment()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 4.2898413],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.6352585],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 5.6451702],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_calculate_captital_investment_empty_dr_idv(
        self,
        capital_cost,
        new_capacity,
        operational_life,
        region,
        year,
        discount_rate,
        discount_rate_idv_empty,
    ):

        results = {
            "CapitalCost": capital_cost,
            "NewCapacity": new_capacity,
            "OperationalLife": operational_life,
            "REGION": region,
            "YEAR": year,
            "DiscountRate": discount_rate,
            "DiscountRateIdv": discount_rate_idv_empty,
        }

        package = ResultsPackage(results)
        actual = package.capital_investment()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 4.104],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.599],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 5.52],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_calculate_captital_investment_no_dr_idv(
        self,
        capital_cost,
        new_capacity,
        operational_life,
        region,
        year,
        discount_rate,
    ):

        results = {
            "CapitalCost": capital_cost,
            "NewCapacity": new_capacity,
            "OperationalLife": operational_life,
            "REGION": region,
            "YEAR": year,
            "DiscountRate": discount_rate,
        }

        package = ResultsPackage(results)
        actual = package.capital_investment()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 4.104],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 1.599],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 5.52],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.capital_investment()
        assert "Cannot calculate CapitalInvestment due to missing data" in str(ex)


class TestCapitalInvestmentStorage:
    def test_capital_investment_storage(
        self, region, year, capital_cost_storage, new_storage_capacity
    ):

        results = {
            "REGION": region,
            "YEAR": year,
            "CapitalCostStorage": capital_cost_storage,
            "NewStorageCapacity": new_storage_capacity,
        }

        package = ResultsPackage(results)
        actual = package.capital_investment_storage()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "BATTERY", 2014, 4.104],
                ["SIMPLICITY", "DAM", 2014, 1.599],
                ["SIMPLICITY", "DAM", 2016, 5.52],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.capital_investment_storage()
        assert "Cannot calculate CapitalInvestmentStorage due to missing data" in str(
            ex
        )


class TestDiscountedCapitalInvestment:
    def test_calculate_discounted_captital_investment(
        self,
        region,
        year,
        undiscounted_capital_investment,
        discount_rate,
    ):

        results = {
            "REGION": region,
            "YEAR": year,
            "DiscountRate": discount_rate,
            "CapitalInvestment": undiscounted_capital_investment,
        }

        package = ResultsPackage(results)
        actual = package.discounted_capital_investment()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 10],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 123],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 434.28571428],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 715.64625850],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.discounted_capital_investment()
        assert (
            "Cannot calculate DiscountedCapitalInvestment due to missing data"
            in str(ex)
        )


class TestDiscountedCapitalInvestmentStorage:
    def test_calculate_discounted_captital_investment_storage(
        self,
        region,
        year,
        undiscounted_capital_investment_storage,
        discount_rate_storage,
    ):

        results = {
            "REGION": region,
            "YEAR": year,
            "DiscountRateStorage": discount_rate_storage,
            "CapitalInvestmentStorage": undiscounted_capital_investment_storage,
        }

        package = ResultsPackage(results)
        actual = package.discounted_capital_investment_storage()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DAM", 2014, 1.23],
                ["SIMPLICITY", "DAM", 2015, 2.22857143],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.discounted_capital_investment_storage()
        assert (
            "Cannot calculate DiscountedCapitalInvestmentStorage due to missing data"
            in str(ex)
        )


class TestDiscountedOperationalCost:
    def test_calculate_discounted_operational_cost(
        self,
        region,
        year,
        discount_rate,
        annual_fixed_operating_cost,
        annual_variable_operating_cost,
    ):

        results = {
            "REGION": region,
            "YEAR": year,
            "DiscountRate": discount_rate,
            "AnnualFixedOperatingCost": annual_fixed_operating_cost,
            "AnnualVariableOperatingCost": annual_variable_operating_cost,
        }

        package = ResultsPackage(results)
        actual = package.discounted_operational_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 19.51800146],
                ["SIMPLICITY", "DUMMY", 2015, 9.29428640],
                ["SIMPLICITY", "DUMMY", 2016, 8.85170134],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, 433.29963238],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, 1031.66579140],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1572.06215832],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.discounted_operational_cost()
        assert "Cannot calculate DiscountedOperationalCost due to missing data" in str(
            ex
        )


class TestDiscountedCostByTechnology:
    def test_calculate_discounted_cost_by_technology(
        self,
        discounted_capital_costs,
        discounted_operational_costs,
        discounted_emissions_penalty,
        discounted_salvage_value,
    ):

        results = {
            "DiscountedCapitalInvestment": discounted_capital_costs,
            "DiscountedOperationalCost": discounted_operational_costs,
            "DiscountedTechnologyEmissionsPenalty": discounted_emissions_penalty,
            "DiscountedSalvageValue": discounted_salvage_value,
        }

        package = ResultsPackage(results)
        actual = package.discounted_technology_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DUMMY", 2014, 24.0],
                ["SIMPLICITY", "DUMMY", 2015, 8.0],
                ["SIMPLICITY", "DUMMY", 2016, 17.0],
                ["SIMPLICITY", "GAS_EXTRACTION", 2014, -333.0],
                ["SIMPLICITY", "GAS_EXTRACTION", 2015, -222.0],
                ["SIMPLICITY", "GAS_EXTRACTION", 2016, 1775.0],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.discounted_technology_cost()
        assert "Cannot calculate DiscountedCostByTechnology due to missing data" in str(
            ex
        )


class TestDiscountedCostByStorage:
    def test_calculate_discounted_cost_by_storage(
        self,
        discounted_capital_costs_storage,
        discounted_salvage_value_storage,
    ):

        results = {
            "DiscountedCapitalInvestmentStorage": discounted_capital_costs_storage,
            "DiscountedSalvageValueStorage": discounted_salvage_value_storage,
        }

        package = ResultsPackage(results)
        actual = package.discounted_storage_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "BATTERY", 2014, 6.54],
                ["SIMPLICITY", "BATTERY", 2015, -5.67],
                ["SIMPLICITY", "BATTERY", 2016, 15.42],
                ["SIMPLICITY", "DAM", 2014, 32.07],
                ["SIMPLICITY", "DAM", 2015, 42.06],
                ["SIMPLICITY", "DAM", 2016, 52.05],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.discounted_storage_cost()
        assert (
            "Cannot calculate TotalDiscountedCostByStorage due to missing data"
            in str(ex)
        )


class TestTotalDiscountedCost:
    def test_calculate_total_discounted_cost(
        self, discounted_technology_cost, discounted_storage_cost
    ):

        results = {
            "DiscountedCostByTechnology": discounted_technology_cost,
            "DiscountedCostByStorage": discounted_storage_cost,
        }

        package = ResultsPackage(results)
        actual = package.total_discounted_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", 2014, 610.5],
                ["SIMPLICITY", 2015, 854.7],
                ["SIMPLICITY", 2016, 1098.9],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_calculate_total_discounted_cost_no_storage(
        self, discounted_technology_cost
    ):
        """Situations where NewStorageCapacity not available"""

        results = {
            "DiscountedCostByTechnology": discounted_technology_cost,
        }

        package = ResultsPackage(results)
        actual = package.total_discounted_cost()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", 2014, 555.0],
                ["SIMPLICITY", 2015, 777.0],
                ["SIMPLICITY", 2016, 999.0],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.total_discounted_cost()
        assert "Cannot calculate TotalDiscountedCost due to missing data" in str(ex)


class TestDiscountedSalvageValueStorage:
    def test_calculate_discounted_salvage_value_storage(
        self, region, year, storage, salvage_value_storage, discount_rate_storage
    ):

        results = {
            "REGION": region,
            "YEAR": year,
            "STORAGE": storage,
            "DiscountRateStorage": discount_rate_storage,
            "SalvageValueStorage": salvage_value_storage,
        }

        package = ResultsPackage(results)
        actual = package.discounted_salvage_value_storage()
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DAM", 2014, 0.87413804],
                ["SIMPLICITY", "DAM", 2015, 1.66299431],
                ["SIMPLICITY", "DAM", 2016, 2.45185059],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_null(self, null: ResultsPackage):
        """ """
        package = null
        with raises(KeyError) as ex:
            package.discounted_salvage_value_storage()
        assert (
            "Cannot calculate DiscountedSalvageValueStorage due to missing data"
            in str(ex)
        )


class TestCapitalRecoveryFactor:
    def test_crf(self, region, discount_rate_idv, operational_life):

        technologies = ["GAS_EXTRACTION", "DUMMY"]
        regions = region["VALUE"].to_list()
        actual = capital_recovery_factor(
            regions, technologies, discount_rate_idv, operational_life
        )

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 0.523809523],
                ["SIMPLICITY", "DUMMY", 0.365558912],
            ],
            columns=["REGION", "TECHNOLOGY", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY"])

        assert_frame_equal(actual, expected)

    def test_crf_null(self, discount_rate_idv, operational_life):

        actual = capital_recovery_factor([], [], discount_rate_idv, operational_life)

        expected = pd.DataFrame(
            data=[],
            columns=["REGION", "TECHNOLOGY", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY"])

        assert_frame_equal(actual, expected)

    def test_crf_no_tech_discount_rate(self, region, discount_rate, operational_life):

        technologies = ["GAS_EXTRACTION", "DUMMY"]
        regions = region["VALUE"].to_list()
        actual = capital_recovery_factor(
            regions, technologies, discount_rate, operational_life
        )

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 0.5121951219512197],
                ["SIMPLICITY", "DUMMY", 0.34972244250594786],
            ],
            columns=["REGION", "TECHNOLOGY", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY"])

        assert_frame_equal(actual, expected)

    def test_crf_empty_discount_rate(
        self, region, discount_rate_empty, operational_life
    ):
        technologies = ["GAS_EXTRACTION", "DUMMY"]
        regions = region["VALUE"].to_list()

        with raises(ValueError):
            capital_recovery_factor(
                regions, technologies, discount_rate_empty, operational_life
            )


class TestPvAnnuity:
    def test_pva(self, region, discount_rate, operational_life):

        technologies = ["GAS_EXTRACTION", "DUMMY"]
        regions = region["VALUE"].to_list()
        actual = pv_annuity(regions, technologies, discount_rate, operational_life)

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "GAS_EXTRACTION", 1.952380952],
                ["SIMPLICITY", "DUMMY", 2.859410430],
            ],
            columns=["REGION", "TECHNOLOGY", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY"])

        assert_frame_equal(actual, expected)

    def test_pva_null(self, discount_rate, operational_life):

        actual = pv_annuity([], [], discount_rate, operational_life)

        expected = pd.DataFrame(
            data=[],
            columns=["REGION", "TECHNOLOGY", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY"])

        assert_frame_equal(actual, expected)

    def test_pva_empty_discount_rate(
        self, region, discount_rate_empty, operational_life
    ):
        technologies = ["GAS_EXTRACTION", "DUMMY"]
        regions = region["VALUE"].to_list()

        with raises(ValueError):
            pv_annuity(regions, technologies, discount_rate_empty, operational_life)


class TestDiscountFactor:
    def test_df_start(self, region, year, discount_rate):

        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor(regions, years, discount_rate, 0.0)

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", 2014, 1],
                ["SIMPLICITY", 2015, 1.05],
                ["SIMPLICITY", 2016, 1.1025],
                ["SIMPLICITY", 2017, 1.157625],
                ["SIMPLICITY", 2018, 1.21550625],
                ["SIMPLICITY", 2019, 1.276281562],
                ["SIMPLICITY", 2020, 1.340095640],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_mid(self, region, year, discount_rate):

        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor(regions, years, discount_rate, 0.5)

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", 2014, 1.024695076],
                ["SIMPLICITY", 2015, 1.075929830],
                ["SIMPLICITY", 2016, 1.129726321],
                ["SIMPLICITY", 2017, 1.186212638],
                ["SIMPLICITY", 2018, 1.245523269],
                ["SIMPLICITY", 2019, 1.307799433],
                ["SIMPLICITY", 2020, 1.373189405],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_end(self, region, year, discount_rate):

        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor(regions, years, discount_rate, 1.0)

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", 2014, 1.05],
                ["SIMPLICITY", 2015, 1.1025],
                ["SIMPLICITY", 2016, 1.157625],
                ["SIMPLICITY", 2017, 1.21550625],
                ["SIMPLICITY", 2018, 1.276281562],
                ["SIMPLICITY", 2019, 1.340095640],
                ["SIMPLICITY", 2020, 1.407100422],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_null(self, discount_rate):

        actual = discount_factor([], [], discount_rate, 0.0)

        expected = pd.DataFrame(
            data=[],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_empty_discount_rate(self, region, year, discount_rate_empty):
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()

        with raises(ValueError):
            discount_factor(regions, years, discount_rate_empty, 1.0)

    def test_df_two_regions(
        self, region_multiple, year, discount_rate_multiple_regions
    ):

        regions = region_multiple["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor(regions, years, discount_rate_multiple_regions, 0.0)

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", 2014, 1],
                ["SIMPLICITY", 2015, 1.05],
                ["SIMPLICITY", 2016, 1.1025],
                ["SIMPLICITY", 2017, 1.157625],
                ["SIMPLICITY", 2018, 1.21550625],
                ["SIMPLICITY", 2019, 1.276281562],
                ["SIMPLICITY", 2020, 1.340095640],
                ["DUMMY", 2014, 1],
                ["DUMMY", 2015, 1.05],
                ["DUMMY", 2016, 1.1025],
                ["DUMMY", 2017, 1.157625],
                ["DUMMY", 2018, 1.21550625],
                ["DUMMY", 2019, 1.276281562],
                ["DUMMY", 2020, 1.340095640],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])

        assert_frame_equal(actual, expected)


class TestDiscountFactorStorage:
    def test_dfs_start(self, region, year, discount_rate_storage):

        storages = ["DAM"]
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor_storage(
            regions, storages, years, discount_rate_storage, 0.0
        )

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DAM", 2014, 1],
                ["SIMPLICITY", "DAM", 2015, 1.05],
                ["SIMPLICITY", "DAM", 2016, 1.1025],
                ["SIMPLICITY", "DAM", 2017, 1.157625],
                ["SIMPLICITY", "DAM", 2018, 1.21550625],
                ["SIMPLICITY", "DAM", 2019, 1.276281562],
                ["SIMPLICITY", "DAM", 2020, 1.340095640],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_dfs_mid(self, region, year, discount_rate_storage):

        storages = ["DAM"]
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor_storage(
            regions, storages, years, discount_rate_storage, 0.5
        )

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DAM", 2014, 1.024695076],
                ["SIMPLICITY", "DAM", 2015, 1.075929830],
                ["SIMPLICITY", "DAM", 2016, 1.129726321],
                ["SIMPLICITY", "DAM", 2017, 1.186212638],
                ["SIMPLICITY", "DAM", 2018, 1.245523269],
                ["SIMPLICITY", "DAM", 2019, 1.307799433],
                ["SIMPLICITY", "DAM", 2020, 1.373189405],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_dfs_end(self, region, year, discount_rate_storage):

        storages = ["DAM"]
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor_storage(
            regions, storages, years, discount_rate_storage, 1.0
        )

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DAM", 2014, 1.05],
                ["SIMPLICITY", "DAM", 2015, 1.1025],
                ["SIMPLICITY", "DAM", 2016, 1.157625],
                ["SIMPLICITY", "DAM", 2017, 1.21550625],
                ["SIMPLICITY", "DAM", 2018, 1.276281562],
                ["SIMPLICITY", "DAM", 2019, 1.340095640],
                ["SIMPLICITY", "DAM", 2020, 1.407100422],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_null(self, discount_rate_storage):

        actual = discount_factor_storage([], [], [], discount_rate_storage, 0.0)

        expected = pd.DataFrame(
            data=[],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_storage_empty_discount_rate(
        self, region, year, discount_rate_storage_empty
    ):
        storages = ["DAM"]
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()

        with raises(ValueError):
            discount_factor_storage(
                regions, storages, years, discount_rate_storage_empty, 1.0
            )


class TestDiscountFactorStorageSalvage:
    def test_discount_factor_storage_salvage(self, region, year, discount_rate_storage):

        storages = ["DAM"]
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()
        actual = discount_factor_storage_salvage(
            regions, storages, years, discount_rate_storage
        )

        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "DAM", 2014, 1.40710042],
                ["SIMPLICITY", "DAM", 2015, 1.40710042],
                ["SIMPLICITY", "DAM", 2016, 1.40710042],
                ["SIMPLICITY", "DAM", 2017, 1.40710042],
                ["SIMPLICITY", "DAM", 2018, 1.40710042],
                ["SIMPLICITY", "DAM", 2019, 1.40710042],
                ["SIMPLICITY", "DAM", 2020, 1.40710042],
            ],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_null(self, discount_rate_storage):

        actual = discount_factor_storage_salvage([], [], [], discount_rate_storage)

        expected = pd.DataFrame(
            data=[],
            columns=["REGION", "STORAGE", "YEAR", "VALUE"],
        ).set_index(["REGION", "STORAGE", "YEAR"])

        assert_frame_equal(actual, expected)

    def test_df_storage_empty_discount_rate(
        self, region, year, discount_rate_storage_empty
    ):
        storages = ["DAM"]
        regions = region["VALUE"].to_list()
        years = year["VALUE"].to_list()

        with raises(ValueError):
            discount_factor_storage_salvage(
                regions, storages, years, discount_rate_storage_empty
            )


class TestResultsPackage:
    def test_results_package_init(self):

        package = ResultsPackage({})

        with raises(KeyError):
            package["BlaBla"]

    def test_results_package_dummy_results(self):
        """Access results using dictionary keys"""
        package = ResultsPackage({"BlaBla": pd.DataFrame()})

        pd.testing.assert_frame_equal(package["BlaBla"], pd.DataFrame())



"""Provides a command line interface to the ``otoole`` package

The key functions are **convert**, **results**, **validate**, and **viz**.

The ``convert`` command allows convertion of multiple different OSeMOSYS input formats
including from/to csv, an AMPL format datafile, a folder of
CSVs, an Excel workbook with one tab per parameter, an SQLite database

The ``results`` command allows conversion of a results solution file into processed
tabular result file(s). Can process results from ``cplex``, ``gurobi``, ``highs``,
``cbc``, and ``glpk``.

The ``validate`` command checks the technology and fuel names
against a standard or user defined configuration file.

The ``viz`` command allows you to produce a Reference Energy System diagram

Example
-------

Ask for help on the command line::

    >>> $ otoole --help
    usage: otoole [-h] [--verbose] [--version] {convert,results,validate,viz} ...

    otoole: Python toolkit of OSeMOSYS users

    positional arguments:
    {convert,cplex,validate,viz}
        convert             Convert from one input format to another
        cplex               Process a CPLEX solution file
        validate            Validate an OSeMOSYS model
        viz                 Visualise the model

    optional arguments:
    -h, --help            show this help message and exit
    --verbose, -v         Enable debug mode
    --version, -V         The version of otoole

"""

import argparse
import logging
import os
import shutil
import sys

from otoole import __version__, convert, convert_results, read
from otoole.exceptions import OtooleSetupError
from otoole.preprocess.setup import get_config_setup_data, get_csv_setup_data
from otoole.utils import read_packaged_file
from otoole.validate import main as validate
from otoole.visualise import create_res
from otoole.write_strategies import WriteCsv

logger = logging.getLogger(__name__)


def validate_model(args):
    data_format = args.data_format
    data_file = args.data_file
    user_config = args.user_config

    input_data, _ = read(user_config, data_format, data_file)

    if args.validate_config:
        validation_config = read_packaged_file(args.validate_config)
        validate(input_data, validation_config)
    else:
        validate(input_data)


def _result_matrix(args):
    """Covert results"""
    convert_results(
        args.config,
        args.from_format,
        args.to_format,
        args.from_path,
        args.to_path,
        args.input_format,
        args.input_path,
        write_defaults=args.write_defaults,
        glpk_model=args.glpk_model,
    )


def _conversion_matrix(args):
    """Convert from one format to another

    Implemented conversion functions::

        from\to     ex cs df
        --------------------
        excel       -- yy --
        csv         nn -- nn
        datafile    nn -- --

    """
    convert(
        args.config,
        args.from_format,
        args.to_format,
        args.from_path,
        args.to_path,
        write_defaults=args.write_defaults,
        keep_whitespace=args.keep_whitespace,
    )


def data2res(args):
    """Get input data and call res creation."""

    data_format = args.data_format
    data_path = args.data_path
    config = args.config
    resfile = args.resfile

    input_data, _ = read(config, data_format, data_path)

    create_res(input_data, resfile)


def setup(args):
    """Creates template data"""

    data_type = args.data_type
    data_path = args.data_path
    overwrite = args.overwrite

    if os.path.exists(data_path) and not overwrite:
        raise OtooleSetupError(resource=data_path)

    if data_type == "config":
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "preprocess", "config.yaml"),
            data_path,
        )
    elif data_type == "csv":
        config = get_config_setup_data()
        input_data, default_values = get_csv_setup_data(config)
        WriteCsv(user_config=config).write(input_data, data_path, default_values)


def get_parser():
    parser = argparse.ArgumentParser(
        description="otoole: Python toolkit of OSeMOSYS users"
    )

    parser.add_argument(
        "--verbose", "-v", help="Enable debug mode", action="count", default=0
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=__version__,
        help="The version of otoole",
    )

    subparsers = parser.add_subparsers()

    # Parser for results
    result_parser = subparsers.add_parser("results", help="Post-process solution files")
    result_parser.add_argument(
        "from_format",
        help="Result data format to convert from",
        choices=sorted(["cbc", "cplex", "gurobi", "glpk", "highs"]),
    )
    result_parser.add_argument(
        "to_format",
        help="Result data format to convert to",
        choices=sorted(["csv", "excel"]),
    )
    result_parser.add_argument("from_path", help="Path to solver solution file")
    result_parser.add_argument("to_path", help="Path to file or folder to convert to")

    result_parser.add_argument(
        "input_format",
        help="Input data format",
        choices=sorted(["csv", "datafile", "excel"]),
    )
    result_parser.add_argument("input_path", help="Path to input_data")
    result_parser.add_argument("config", help="Path to config YAML file")
    result_parser.add_argument(
        "--glpk_model",
        help="GLPK model file required for processing GLPK results",
        default=None,
    )
    result_parser.add_argument(
        "--write_defaults",
        help="Writes default values",
        default=False,
        action="store_true",
    )
    result_parser.set_defaults(func=_result_matrix)

    # Parser for conversion
    convert_parser = subparsers.add_parser(
        "convert", help="Convert from one input format to another"
    )
    convert_parser.add_argument(
        "from_format",
        help="Input data format to convert from",
        choices=sorted(["datafile", "excel", "csv"]),
    )
    convert_parser.add_argument(
        "to_format",
        help="Input data format to convert to",
        choices=sorted(["datafile", "csv", "excel"]),
    )
    convert_parser.add_argument(
        "from_path", help="Path to file or folder to convert from"
    )
    convert_parser.add_argument("to_path", help="Path to file or folder to convert to")
    convert_parser.add_argument("config", help="Path to config YAML file")
    convert_parser.add_argument(
        "--write_defaults",
        help="Writes default values",
        default=False,
        action="store_true",
    )
    convert_parser.add_argument(
        "--keep_whitespace",
        help="Keeps leading/trailing whitespace",
        default=False,
        action="store_true",
    )
    convert_parser.set_defaults(func=_conversion_matrix)

    # Parser for validation
    valid_parser = subparsers.add_parser("validate", help="Validate an OSeMOSYS model")
    valid_parser.add_argument(
        "data_format",
        help="Input data format",
        choices=sorted(["datafile", "excel", "csv"]),
    )
    valid_parser.add_argument(
        "data_file", help="Path to the OSeMOSYS data file to validate"
    )
    valid_parser.add_argument("user_config", help="Path to config YAML file")
    valid_parser.add_argument(
        "--validate_config", help="Path to a user-defined validation-config file"
    )
    valid_parser.set_defaults(func=validate_model)

    # Parser for visualisation
    viz_parser = subparsers.add_parser("viz", help="Visualise the model")
    viz_subparsers = viz_parser.add_subparsers()

    res_parser = viz_subparsers.add_parser(
        "res", help="Generate a reference energy system"
    )
    res_parser.add_argument(
        "data_format",
        help="Input data format",
        choices=sorted(["datafile", "excel", "csv"]),
    )
    res_parser.add_argument("data_path", help="Input data path")
    res_parser.add_argument("resfile", help="Path to reference energy system")
    res_parser.add_argument("config", help="Path to config YAML file")
    res_parser.set_defaults(func=data2res)

    # parser for setup
    setup_parser = subparsers.add_parser("setup", help="Setup template files")
    setup_parser.add_argument(
        "data_type", help="Type of file to setup", choices=sorted(["config", "csv"])
    )
    setup_parser.add_argument("data_path", help="Path to file or folder to save to")
    setup_parser.add_argument(
        "--overwrite",
        help="Overwrites existing data",
        default=False,
        action="store_true",
    )
    setup_parser.set_defaults(func=setup)

    return parser


class DeprecateAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        logger.warning(f"Argument {self.option_strings} is deprecated and is ignored.")
        delattr(namespace, self.dest)


def main():

    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])

    if args.verbose >= 1 and args.verbose < 2:
        logging.basicConfig(level=logging.INFO)
    if args.verbose > 2:
        logging.basicConfig(level=logging.DEBUG)

    def exception_handler(
        exception_type, exception, traceback, debug_hook=sys.excepthook
    ):
        if args.verbose:
            debug_hook(exception_type, exception, traceback)
        else:
            print("{}: {}".format(exception_type.__name__, str(exception)))

    sys.excepthook = exception_handler

    if "func" in args:
        args.func(args)
    else:
        parser.print_help()



from typing import List, Union


class OtooleException(Exception):
    """Base class for all otoole exceptions."""

    pass


class OtooleValidationError(OtooleException):
    """Input data is invalid

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, message):
        self.resource = resource
        self.message = message


class OtooleRelationError(OtooleException):
    """Relations between input data is not correct

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    foreign_resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, foreign_resource, message):
        self.resource = resource
        self.foreign_resource = foreign_resource
        self.message = message


class OtooleConfigFileError(OtooleException):
    """Config file validation error

    Arguments
    ---------
    message: str
        Message to display to users
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self):
        return f"{self.message}"


class OtooleExcelNameLengthError(OtooleException):
    """Invalid tab name for writing to Excel."""

    def __init__(
        self,
        name: str,
        message: str = "Parameter name must be less than 31 characters when writing to Excel",
    ) -> None:
        self.name = name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.name} -> {self.message}"


class OtooleNameMismatchError(OtooleException):
    """Names not consistent between read in data and config file"""

    def __init__(
        self,
        name: Union[List, str],
    ) -> None:
        if isinstance(name, list):
            self.message = "Names not consistent between data and config file"
            self.name = ", ".join(sorted(name))
        elif isinstance(name, str):
            self.name = name
            self.message = "Name not consistent between data and config file"
        else:
            raise TypeError("Incorrect type passed to OtooleNameMismatchError")
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}:\n\n{self.name}.\n\nUpdate config or data with matching names."


class OtooleDeprecationError(OtooleException):
    """New version of otoole does drops this feature support

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, message):
        self.resource = resource
        self.message = message

    def __str__(self):
        return f"{self.resource} -> {self.message}"


class OtooleSetupError(OtooleException):
    """Setup data already exists

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(
        self,
        resource,
        message="Data already exists. Delete file/directory or pass the --overwrite flag",
    ):
        self.resource = resource
        self.message = message

    def __str__(self):
        return f"{self.resource} -> {self.message}"


class OtooleIndexError(OtooleException):
    """Index data not consistent between data and config file

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    config_indices: List[str]
        Indices from config file
    data_indices: List[str]
        Indices from input data
    """

    def __init__(self, resource, config_indices, data_indices):
        self.resource = resource
        self.config_indices = config_indices
        self.data_indices = data_indices
        self.message = "Indices inconsistent between config and data"

    def __str__(self):
        return f"{self.resource} -> {self.message}. Config indices are {self.config_indices}. Data indices are {self.data_indices}."


class OtooleError(OtooleException):
    """General purpose error

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, message):
        self.resource = resource
        self.message = message

    def __str__(self):
        return f"{self.resource} -> {self.message}"



import logging
import os
from typing import Any, TextIO

import pandas as pd

from otoole.exceptions import OtooleExcelNameLengthError
from otoole.input import WriteStrategy

logger = logging.getLogger(__name__)


class WriteExcel(WriteStrategy):
    def _header(self):
        return pd.ExcelWriter(self.filepath, mode="w")

    def _form_parameter(
        self, df: pd.DataFrame, parameter_name: str, default: float
    ) -> pd.DataFrame:
        """Converts data into wide format

        Arguments
        ---------
        df: pd.DataFrame
        parameter_name: str
        default: float

        Returns
        -------
        pandas.DataFrame
        """

        index_names = df.index.names
        column_names = df.columns.to_list()
        if index_names[0]:
            names = index_names + column_names
        else:
            names = column_names
        logger.debug(f"Identified {len(names)} names: {names}")

        total_columns = len(names)

        if "YEAR" not in names:
            pivot = df.copy()
        elif total_columns > 3:
            logger.debug("More than 3 columns for {}: {}".format(parameter_name, names))
            rows = names[0:-2]
            columns = names[-2]
            values = names[-1]
            logger.debug(f"Rows: {rows}; columns: {columns}; values: {values}")
            logger.debug("dtypes: {}".format(df.dtypes))
            pivot = df.reset_index().pivot(index=rows, columns=columns, values=values)
        else:
            logger.debug(f"One column for {parameter_name}: {names}")
            pivot = df.copy()

        return pivot

    def _form_parameter_template(self, parameter_name: str, **kwargs) -> pd.DataFrame:
        """Creates wide format excel template

        Pivots the data to wide format using the data from the YEAR set as the columns.
        This requires input data to be passed into this function.

        Arguments
        ---------
        parameter_name: str
        input_data: dict[str, pd.DataFrame])

        Returns
        -------
        pd.DataFrame
        """

        indices = self.user_config[parameter_name]["indices"]

        if "input_data" not in kwargs:
            logger.debug(f"Can not pivot excel template for {parameter_name}")
            return pd.DataFrame(columns=indices)
        else:
            input_data = kwargs["input_data"]

        if "YEAR" in indices:
            years = input_data["YEAR"]["VALUE"].to_list()
            indices.remove("YEAR")
            indices.extend(years)
        else:
            indices.extend(["VALUE"])

        return pd.DataFrame(columns=indices)

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: pd.ExcelWriter,
        default: float,
        **kwargs,
    ):
        try:
            name = self.user_config[parameter_name]["short_name"]
        except KeyError:
            name = parameter_name

        if len(name) > 31:
            raise OtooleExcelNameLengthError(name=name)

        if not df.empty:
            df = self._form_parameter(df, parameter_name, default)
            df.to_excel(handle, sheet_name=name, merge_cells=False, index=True)
        else:
            logger.debug(f"Dataframe {parameter_name} is empty")
            df = self._form_parameter_template(parameter_name, **kwargs)
            df.to_excel(handle, sheet_name=name, merge_cells=False, index=False)

    def _write_set(self, df: pd.DataFrame, set_name, handle: pd.ExcelWriter):
        df.to_excel(handle, sheet_name=set_name, merge_cells=False, index=False)

    def _footer(self, handle=pd.ExcelWriter):
        handle.close()


class WriteDatafile(WriteStrategy):
    def _header(self):
        filepath = open(self.filepath, "w", newline="")
        msg = "# Model file written by *otoole*\n"
        filepath.write(msg)
        return filepath

    def _form_parameter(self, df: pd.DataFrame, default: float):

        # Don't write out values equal to the default value
        df = df[df.VALUE != default]
        return df

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: TextIO,
        default: float,
        **kwargs,
    ):
        """Write parameter data to a GMPL datafile, omitting data with default value

        Arguments
        ---------
        filepath : StreamIO
        df : pandas.DataFrame
        parameter_name : str
        handle: TextIO
        default : int
        """

        df = self._form_parameter(df, default)
        handle.write("param default {} : {} :=\n".format(default, parameter_name))
        df.to_csv(
            path_or_buf=handle,
            sep=" ",
            header=False,
            index=True,
            float_format="%g",
            lineterminator="\n",
        )
        handle.write(";\n")

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO):
        """Write set data to a GMPL datafile

        Arguments
        ---------
        df : pandas.DataFrame
        set_name : str
        handle: TextIO
        """
        handle.write("set {} :=\n".format(set_name))
        df.to_csv(
            path_or_buf=handle,
            sep=" ",
            header=False,
            index=False,
            float_format="%g",
            lineterminator="\n",
        )
        handle.write(";\n")

    def _footer(self, handle: TextIO):
        handle.write("end;\n")
        handle.close()


class WriteCsv(WriteStrategy):
    """Write parameters to comma-separated value files

    Arguments
    ---------
    filepath: str, default=None
        The path to write a folder of csv files
    default_values: dict, default=None
    user_config: dict, default=None
    """

    @staticmethod
    def _write_out_dataframe(folder, parameter, df, index=False):
        """Writes out a dataframe as a csv into a data subfolder

        Arguments
        ---------
        folder : str
        parameter : str
        df : pandas.DataFrame
        index : bool, default=False
            Write the index to CSV
        """
        filepath = os.path.join(folder, parameter + ".csv")
        with open(filepath, "w", newline="") as csvfile:
            logger.info(
                "Writing %s rows into narrow file for %s", df.shape[0], parameter
            )
            df.to_csv(csvfile, index=index)

    def _header(self) -> Any:
        os.makedirs(os.path.join(self.filepath), exist_ok=True)
        return None

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: TextIO,
        default: float,
        **kwargs,
    ) -> pd.DataFrame:
        """Write parameter data"""
        self._write_out_dataframe(self.filepath, parameter_name, df, index=True)

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        """Write set data"""
        self._write_out_dataframe(self.filepath, set_name, df, index=False)

    def _footer(self, handle: TextIO):
        pass



import logging
from abc import abstractmethod
from io import StringIO
from typing import Any, Dict, TextIO, Tuple, Union

import pandas as pd

from otoole.input import ReadStrategy
from otoole.results.result_package import ResultsPackage

LOGGER = logging.getLogger(__name__)


class ReadResults(ReadStrategy):
    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        """Read a solution file from ``filepath`` and process using ``input_data``

        Arguments
        ---------
        filepath : str, TextIO
            A path name or file buffer pointing to the solution file
        input_data : dict, default=None
            dict of dataframes

        Returns
        -------
        tuple
            A tuple containing dict of pandas.DataFrames and a dict of default_values

        """
        if "input_data" in kwargs:
            input_data = kwargs["input_data"]
            param_default_values = self._read_default_values(self.input_config)
        else:
            input_data = {}
            param_default_values = {}

        available_results = self.get_results_from_file(
            filepath, input_data
        )  # type: Dict[str, pd.DataFrame]

        default_values = self._read_default_values(self.results_config)  # type: Dict

        input_data = self._expand_required_params(input_data, param_default_values)

        results = self.calculate_results(
            available_results, input_data
        )  # type: Dict[str, pd.DataFrame]

        if self.write_defaults:
            results = self.write_default_results(results, input_data, default_values)

        return results, default_values

    @abstractmethod
    def get_results_from_file(self, filepath, input_data):
        raise NotImplementedError()

    def calculate_results(
        self,
        available_results: Dict[str, pd.DataFrame],
        input_data: Dict[str, pd.DataFrame],
    ) -> Dict[str, pd.DataFrame]:
        """Populates the results with calculated values using input data"""

        results = {}
        results_package = ResultsPackage(available_results, input_data)

        for name in sorted(self.results_config.keys()):

            LOGGER.info("Looking for %s", name)

            try:
                results[name] = results_package[name]
            except KeyError as ex:
                LOGGER.info("No calculation method available for %s", name)
                LOGGER.debug("Error calculating %s: %s", name, str(ex))

        return results

    def _expand_required_params(
        self,
        input_data: dict[str, pd.DataFrame],
        param_defaults: dict[str, Any],
    ) -> dict[str, pd.DataFrame]:
        """Expands required default values for results processing"""

        if "DiscountRate" in input_data:
            input_data["DiscountRate"] = self._expand_dataframe(
                "DiscountRate", input_data, param_defaults
            )
        if "DiscountRateIdv" in input_data:
            input_data["DiscountRateIdv"] = self._expand_dataframe(
                "DiscountRateIdv", input_data, param_defaults
            )

        return input_data


class ReadWideResults(ReadResults):
    def get_results_from_file(self, filepath, input_data):
        cbc = self._convert_to_dataframe(filepath)
        available_results = self._convert_wide_to_long(cbc)
        return available_results

    @abstractmethod
    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        raise NotImplementedError()

    def _convert_wide_to_long(self, data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        """Convert from wide to long format

        Converts a pandas DataFrame containing all wide format results to reformatted
        dictionary of pandas DataFrames in long format ready to write out

        Arguments
        ---------
        data : pandas.DataFrame
            results stored in a dataframe

        Example
        -------
        >>> df = pd.DataFrame(data=[
                ['TotalDiscountedCost', "SIMPLICITY,2015", 187.01576],
                ['TotalDiscountedCost', "SIMPLICITY,2016", 183.30788]],
                columns=['Variable', 'Index', 'Value'])
        >>> convert_dataframe_to_csv(df)
        {'TotalDiscountedCost':        REGION  YEAR      VALUE
                                    0  SIMPLICITY  2015  187.01576
                                    1  SIMPLICITY  2016  183.30788}
        """

        sets = {x: y for x, y in self.user_config.items() if y["type"] == "set"}

        results = {}  # type: Dict[str, pd.DataFrame]
        not_found = []

        for name, details in sorted(self.results_config.items()):
            df_cbc = data[data["Variable"] == name]

            if not df_cbc.empty:

                df = df_cbc.copy()  # setting with copy warning
                LOGGER.debug("Extracting results for %s", name)
                indices = details["indices"]  # typing: List

                df[indices] = df["Index"].str.split(",", expand=True)

                types = {index: sets[index]["dtype"] for index in indices}
                df = df.astype(types)

                df = df.drop(columns=["Variable", "Index"])

                df = df.rename(columns={"Value": "VALUE"})

                columns = indices + ["VALUE"]

                df = df[columns]

                index = details["indices"].copy()
                df, index = check_duplicate_index(df, columns, index)
                results[name] = df.set_index(index)
            else:
                not_found.append(name)

        LOGGER.debug("Unable to find result variables for: %s", ", ".join(not_found))

        return results


def check_duplicate_index(df: pd.DataFrame, columns: list, index: list) -> pd.DataFrame:
    """Catches pandas error when there are duplicate column indices"""
    if check_for_duplicates(index):
        index = rename_duplicate_column(index)
        LOGGER.debug("Original column names: %s", columns)
        renamed_columns = rename_duplicate_column(columns)
        LOGGER.debug("New column names: %s", renamed_columns)
        df.columns = renamed_columns
    return df, index


def check_for_duplicates(index: list) -> bool:
    return len(set(index)) != len(index)


def identify_duplicate(index: list) -> Union[int, bool]:
    elements = set()  # type: set
    for counter, elem in enumerate(index):
        if elem in elements:
            return counter
        else:
            elements.add(elem)
    return False


def rename_duplicate_column(index: list) -> list:
    column = index.copy()
    location = identify_duplicate(column)
    if location:
        column[location] = "_" + column[location]
    return column


class ReadCplex(ReadWideResults):
    """Read a CPLEX solution file into memeory"""

    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a Cplex solution file into a pandas DataFrame

        Arguments
        ---------
        user_config : Dict[str, Dict]
        file_path : Union[str, TextIO]
        """
        df = pd.read_xml(file_path, xpath=".//variable", parser="etree")
        df[["Variable", "Index"]] = df["name"].str.split("(", expand=True)
        df["Index"] = df["Index"].str.replace(")", "", regex=False)
        LOGGER.debug(df)
        df = df[(df["value"] != 0)].reset_index().rename(columns={"value": "Value"})
        return df[["Variable", "Index", "Value"]].astype({"Value": float})


class ReadGurobi(ReadWideResults):
    """Read a Gurobi solution file into memory"""

    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a Gurobi solution file into a pandas DataFrame

        Arguments
        ---------
        user_config : Dict[str, Dict]
        file_path : Union[str, TextIO]
        """
        df = pd.read_csv(
            file_path,
            header=None,
            sep=" ",
            names=["Variable", "Value"],
            skiprows=2,
        )  # type: pd.DataFrame
        df[["Variable", "Index"]] = df["Variable"].str.split("(", expand=True)
        df["Index"] = df["Index"].str.replace(")", "", regex=False)
        LOGGER.debug(df)
        df = df[(df["Value"] != 0)].reset_index()
        return df[["Variable", "Index", "Value"]].astype({"Value": float})


class ReadCbc(ReadWideResults):
    """Read a CBC solution file into memory"""

    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a CBC solution file into a pandas DataFrame

        Arguments
        ---------
        file_path : str
        """
        df = pd.read_csv(
            file_path,
            header=None,
            sep="(",
            names=["Variable", "indexvalue"],
            skiprows=1,
        )  # type: pd.DataFrame
        if df["Variable"].astype(str).str.contains(r"^\*\*").any():
            LOGGER.warning(
                "CBC Solution File contains decision variables out of bounds. "
                + "You have an infeasible solution"
            )
        df["Variable"] = (
            df["Variable"]
            .astype(str)
            .str.replace(r"^\*\*", "", regex=True)
            .str.split(expand=True)[1]
        )
        df[["Index", "Value"]] = df["indexvalue"].str.split(expand=True).loc[:, 0:1]
        df["Index"] = df["Index"].str.replace(")", "", regex=False)
        df = df.drop(columns=["indexvalue"])
        return df[["Variable", "Index", "Value"]].astype({"Value": float})


class ReadHighs(ReadWideResults):
    """Read a HiGHS solution file into memory"""

    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a HiGHS solution file into a pandas DataFrame

        Arguments
        ---------
        file_path : str
        """
        df = pd.read_csv(
            file_path,
            sep=r"\s+",
            skiprows=1,
            index_col=0,
            dtype=str,
        )

        # Type column is not garunteed in the the model output
        # retain conditional as filtering on type is more explicit
        if "Type" in df.columns:
            var_types = ["Continuous", "Integer", "SemiContinuous", "SemiInteger"]
            df = df[df.Type.isin(var_types)].copy()
        else:
            df = df.reset_index()
            row = df[df.Index == "Rows"].index[0]
            df = df.iloc[:row].set_index("Index")

        df.index.name = ""  # remove the name Index, as otoole uses that

        df[["Variable", "Index"]] = df["Name"].str.split("(", expand=True).loc[:, 0:1]
        df["Index"] = df["Index"].str.replace(")", "", regex=False)
        df = df[~(df.Primal.astype(float).abs() < 1e-6)]
        return (
            df[["Variable", "Index", "Primal"]]
            .rename(columns={"Primal": "Value"})
            .reset_index(drop=True)
            .astype({"Variable": str, "Index": str, "Value": float})
        )


class ReadGlpk(ReadWideResults):
    """Reads a GLPK Solution file into memory

    Arguments
    ---------
    user_config : Dict[str, Dict]
    glpk_model: Union[str, TextIO]
        Path to GLPK model file. Can be created using the `--wglp` flag.
    """

    def __init__(
        self,
        user_config: Dict[str, Dict],
        glpk_model: Union[str, TextIO],
        write_defaults: bool = False,
    ):
        super().__init__(user_config=user_config, write_defaults=write_defaults)

        if isinstance(glpk_model, str):
            with open(glpk_model, "r") as model_file:
                self.model = self.read_model(model_file)
        elif isinstance(glpk_model, StringIO):
            self.model = self.read_model(glpk_model)
        else:
            raise TypeError("Argument filepath type must be a string or an open file")

    def _convert_to_dataframe(self, glpk_sol: Union[str, TextIO]) -> pd.DataFrame:
        """Creates a wide formatted dataframe from GLPK solution

        Arguments
        ---------
        glpk_sol: Union[str, TextIO]
            Path to GLPK solution file. Can be created using the `--write` flag

        Returns
        -------
        pd.DataFrame
        """

        if isinstance(glpk_sol, str):
            with open(glpk_sol, "r"):
                _, sol = self.read_solution(glpk_sol)
        elif isinstance(glpk_sol, StringIO):
            _, sol = self.read_solution(glpk_sol)
        else:
            raise TypeError("Argument filepath type must be a string or an open file")

        return self._merge_model_sol(sol)

    def read_model(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads in a GLPK Model File

        Arguments
        ---------
        file_path: Union[str, TextIO]
            Path to GLPK model file. Can be created using the `--wglp` flag.

        Returns
        -------
        pd.DataFrame

           ID  NUM  NAME                      INDEX
        0  i   1    CAa4_Constraint_Capacity  "SIMPLICITY,ID,BACKSTOP1,2015"
        1  j   2    NewCapacity               "SIMPLICITY,WINDPOWER,2039"

        Notes
        -----

        -> GENERAL LAYOUT OF SOLUTION FILE

        n p NAME # p = problem instance
        n z NAME # z = objective function
        n i ROW NAME # i = constraint name, ROW is the row ordinal number
        n j COL NAME # j = variable name, COL is the column ordinal number
        """

        df = pd.read_csv(
            file_path,
            header=None,
            sep=r"\s+",
            index_col=0,
            names=["ID", "NUM", "value", 4, 5],
        ).drop(columns=[4, 5])

        df = df[(df["ID"].isin(["i", "j"])) & (df["value"] != "cost")]

        df[["NAME", "INDEX"]] = df["value"].str.split("[", expand=True)
        df["INDEX"] = df["INDEX"].map(lambda x: x.split("]")[0])
        df = (
            df[["ID", "NUM", "NAME", "INDEX"]]
            .astype({"ID": str, "NUM": "int64", "NAME": str, "INDEX": str})
            .reset_index(drop=True)
        )

        return df

    def read_solution(
        self, file_path: Union[str, TextIO]
    ) -> Tuple[Dict[str, Union[str, float]], pd.DataFrame]:
        """Reads a GLPK solution file

        Arguments
        ---------
        file_path: Union[str, TextIO]
            Path to GLPK solution file. Can be created using the `--write` flag

        Returns
        -------
        Tuple[Dict[str,Union[str, float]], pd.DataFrame]
            Dict[str,Union[str, float]] -> Problem name, status, and objective value
            pd.DataFrame -> Variables and constraints

        {"name":"osemosys", "status":"OPTIMAL", "objective":4497.31976}

           ID  NUM  STATUS PRIM DUAL
        0  i   1    b      5    0
        1  j   2    l      0    2

        Notes
        -----

        -> ROWS IN SOLUTION FILE

        i ROW ST PRIM DUAL

        ROW is the ordinal number of the row
        ST is one of:
        - b = inactive constraint;
        - l = inequality constraint active on its lower bound;
        - u = inequality constraint active on its upper bound;
        - f = active free (unounded) row;
        - s = active equality constraint.
        PRIM specifies the row primal value (float)
        DUAL specifies the row dual value (float)

        -> COLUMNS IN SOLUTION FILE

        j COL ST PRIM DUAL

        COL specifies the column ordinal number
        ST contains one of the following lower-case letters that specifies the column status in the basic solution:
        - b = basic variable
        - l = non-basic variable having its lower bound active
        - u = non-basic variable having its upper bound active
        - f = non-basic free (unbounded) variable
        - s = non-basic fixed variable.
        PRIM field contains column primal value (float)
        DUAL field contains the column dual value (float)
        """

        df = pd.read_csv(file_path, header=None, sep=":")

        # get status information
        status = {}
        df_status = df.loc[:8].set_index(0)
        status["name"] = df_status.loc["c Problem", 1].strip()
        status["status"] = df_status.loc["c Status", 1].strip()
        status["objective"] = float(df_status.loc["c Objective", 1].split()[2])

        # get solution infromation
        data = df.iloc[8:-1].copy()
        data[["ID", "NUM", "STATUS", "PRIM", "DUAL"]] = data[0].str.split(
            " ", expand=True
        )

        data = (
            data[["ID", "NUM", "STATUS", "PRIM", "DUAL"]]
            .astype(
                {"ID": str, "NUM": "int64", "STATUS": str, "PRIM": float, "DUAL": float}
            )
            .reset_index(drop=True)
        )

        return status, data

    def _merge_model_sol(self, sol: pd.DataFrame) -> pd.DataFrame:
        """Merges GLPK model and solution file into one dataframe

        Arguments
        ---------
        sol: pd.DataFrame
            see output from ReadGlpk.read_solution(...)

        Returns
        -------
        pd.DataFrame

        >>> pd.DataFrame(data=[
                ['TotalDiscountedCost', "SIMPLICITY,2015", 187.01576],
                ['TotalDiscountedCost', "SIMPLICITY,2016", 183.30788]],
                columns=['Variable', 'Index', 'Value'])
        """

        model = self.model.copy()
        model.index = model["ID"].str.cat(model["NUM"].astype(str))
        model = model.drop(columns=["ID", "NUM"])

        sol.index = sol["ID"].str.cat(sol["NUM"].astype(str))
        sol = sol.drop(columns=["ID", "NUM", "STATUS", "DUAL"])

        df = model.join(sol)
        df = (
            df[df.index.str.startswith("j")]
            .reset_index(drop=True)
            .rename(columns={"NAME": "Variable", "INDEX": "Index", "PRIM": "Value"})
        )

        return df



import json
import logging
import os
from importlib.resources import files
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pydantic import ValidationError
from yaml import SafeLoader, load  # type: ignore

from otoole.exceptions import OtooleConfigFileError, OtooleDeprecationError
from otoole.preprocess.validate_config import (
    UserDefinedParameter,
    UserDefinedResult,
    UserDefinedSet,
    UserDefinedValue,
)

logger = logging.getLogger(__name__)


def _read_file(open_file, ending):
    if ending == ".yaml" or ending == ".yml":
        contents = load(open_file, Loader=UniqueKeyLoader)  # typing: Dict[str, Any]
    elif ending == ".json":
        contents = json.load(open_file)  # typing: Dict
    else:
        contents = open_file.readlines()
    return contents


def read_packaged_file(filename: str, module_name: Optional[str] = None):

    _, ending = os.path.splitext(filename)

    if module_name is None:
        with open(filename, "r") as open_file:
            contents = _read_file(open_file, ending)
    else:
        with files(module_name).joinpath(filename).open("r") as open_file:
            contents = _read_file(open_file, ending)

    return contents


def extract_config(
    schema: Dict, default_values: Dict
) -> Dict[str, Dict[str, Union[str, List[str]]]]:

    config = {}  # type: Dict[str, Dict[str, Union[str, List[str]]]]
    for resource in schema["resources"]:

        name = resource["name"]
        if name == "default_values":
            continue

        dtype_mapping = {
            "number": "float",
            "string": "str",
            "float": "float",
            "integer": "int",
        }

        fields = resource["schema"]["fields"]
        dtype = [x["type"] for x in fields if x["name"] == "VALUE"][0]
        if (len(fields) == 1) & (fields[0]["name"] == "VALUE"):
            element_type = "set"
            config[name] = {"dtype": dtype_mapping[dtype], "type": element_type}
        else:
            element_type = "param"
            indices = [x["name"] for x in fields if x["name"] != "VALUE"]
            config[name] = {
                "type": element_type,
                "indices": indices,
                "dtype": dtype_mapping[dtype],
                "default": default_values[name],
            }
    return config


def create_name_mappings(
    config: Dict[str, Dict[str, Union[str, List]]], map_full_to_short: bool = True
) -> Dict:
    """Creates name mapping between full name and short name.

    Arguments
    ---------
    config : Dict[str, Dict[str, Union[str, List]]]
        Parsed user configuration file
    map_full_to_short: bool
        Map full name to short name if true, else map short name to full name

    Returns
    -------
    csv_to_excel Dict[str, str]
        Mapping of full name to shortened name

    """

    csv_to_excel = {}
    for name, params in config.items():
        try:
            csv_to_excel[name] = params["short_name"]
        except KeyError:
            if len(name) > 31:
                logger.info(f"{name} does not have a 'short_name'")
            continue

    if map_full_to_short:
        return csv_to_excel
    else:
        return {v: k for k, v in csv_to_excel.items()}


def validate_config(config: Dict) -> None:
    """Validates user input data

    Arguments
    ---------
    config: Dict
        Read in user config yaml file

    Raises
    ------
    ValidationError
        If the user inputs are not valid
    """

    # For validating with pydantic
    config_flattened = format_config_for_validation(config)
    user_defined_sets = get_all_sets(config)

    errors = []
    for input_data in config_flattened:
        try:
            if "type" not in input_data:
                UserDefinedValue(**input_data)
            elif input_data["type"] == "param":
                input_data["defined_sets"] = user_defined_sets
                UserDefinedParameter(**input_data)
            elif input_data["type"] == "result":
                input_data["defined_sets"] = user_defined_sets
                UserDefinedResult(**input_data)
            elif input_data["type"] == "set":
                UserDefinedSet(**input_data)
            else:
                # have pydantic raise an error
                UserDefinedValue(
                    name=input_data["name"],
                    type=input_data["type"],
                    dtype=input_data["dtype"],
                )
        except ValidationError as ex:
            errors_caught = [x["msg"] for x in ex.errors()]
            errors.extend(errors_caught)

    if errors:
        error_message = "\n".join(errors)
        raise OtooleConfigFileError(message=f"\n{error_message}")


def format_config_for_validation(config_in: Dict) -> List:
    """Formats config for validation function.

    Flattens dictionary to a list

    Arguments
    ---------
    config_in: Dict
        Read in user config yaml file

    Returns
    -------
    config_out: List

    Example
    -------
    >>> config_in
    >>> AccumulatedAnnualDemand:
          indices: [REGION,FUEL,YEAR]
          type: param
          dtype: float
          default: 0

    >>> config_out
    >>> [{
        name: AccumulatedAnnualDemand
        indices: [REGION,FUEL,YEAR]
        type: param
        dtype: float
        default: 0
        }, ... ]
    """
    config_out = []
    for name, data in config_in.items():
        flattened_data = {"name": name, **data}
        config_out.append(flattened_data)
    return config_out


def read_deprecated_datapackage(datapackage: str) -> str:
    """Checks filepath for CSVs if a datapackage is provided

    Arguments
    ---------
    datapackage: str
        Location of input datapackge

    Returns
    -------
    input_csvs: str
        Location of input csv files

    Raises
    ------
    OtooleDeprecationError
        If no 'data/' directory is found
    """

    input_csvs = os.path.join(os.path.dirname(datapackage), "data")
    if os.path.exists(input_csvs):
        return input_csvs
    else:
        raise OtooleDeprecationError(
            resource="datapackage.json",
            message="datapackage format no longer supported and no csv data found",
        )


def get_packaged_resource(
    input_data: Dict[str, pd.DataFrame], param: str
) -> List[Dict[str, Any]]:
    """Gets input parameter data and formats it as a dictionary

    Arguments
    ---------
    input_data : Dict[str, pd.DataFrame]
        Internal datastore for otoole input data
    param : str
        Name of OSeMOSYS parameter

    Returns
    -------
    List[Dict[str,any]]
        List of all rows in the df, where each dictionary is the column
        name, followed by the value in that row

    Example
    -------
    >>> get_packaged_resource(input_data, "InputActivityRatio")
    >>> [{'REGION': 'SIMPLICITY',
        'TECHNOLOGY': 'RIVWATAGR',
        'FUEL': 'WATIN',
        'MODE_OF_OPERATION': 1,
        'YEAR': 2020,
        'VALUE': 1.0}]
    """
    return input_data[param].reset_index().to_dict(orient="records")


class UniqueKeyLoader(SafeLoader):
    """YALM Loader to find duplicate keys

    This loader will treat lowercase and uppercase keys as the same. Meaning,
    the keys "SampleKey" and "SAMPLEKEY" are considered the same.

    Raises
    ------
    ValueError
        If a key is defined more than once.

    Adapted from:
    https://stackoverflow.com/a/63215043/14961492
    """

    def construct_mapping(self, node, deep=False):
        mapping = []
        for key_node, _ in node.value:
            key = self.construct_object(key_node, deep=deep)
            key = key.upper()
            if key in mapping:
                raise ValueError(f"{key} -> defined more than once")
            mapping.append(key)
        return super().construct_mapping(node, deep)


def get_all_sets(config: Dict) -> List:
    """Extracts user defined sets"""
    return [x for x, y in config.items() if y["type"] == "set"]



"""This module implements the public API of the otoole package

Use the otoole ``convert`` function to convert between different file formats.
Import the convert function from the otoole package::

>>> from otoole import convert
>>> convert('config.yaml', 'excel', 'datafile', 'input.xlsx', 'output.dat')

"""

import logging
import os
from typing import Dict, Optional, Tuple, Union

import pandas as pd

from otoole.exceptions import OtooleError
from otoole.input import Context, ReadStrategy, WriteStrategy
from otoole.read_strategies import ReadCsv, ReadDatafile, ReadExcel
from otoole.results.results import (
    ReadCbc,
    ReadCplex,
    ReadGlpk,
    ReadGurobi,
    ReadHighs,
    ReadResults,
)
from otoole.utils import _read_file, read_deprecated_datapackage, validate_config
from otoole.write_strategies import WriteCsv, WriteDatafile, WriteExcel

logger = logging.getLogger(__name__)


def read_results(
    config: str,
    from_format: str,
    from_path: str,
    input_format: str,
    input_path: str,
    write_defaults: bool = False,
    glpk_model: Optional[str] = None,
) -> Tuple[Dict[str, pd.DataFrame], Dict[str, float]]:
    """Read OSeMOSYS results from CBC, GLPK, Gurobi, HiGHS, or CPLEX results files

    Arguments
    ---------
    config : str
        Path to config file
    from_format : str
        Available options are 'cbc', 'gurobi', 'cplex', 'highs', and 'glpk'
    from_path : str
        Path to source file (if datafile or excel) or folder (csv)
    input_format: str
        Format of input data. Available options are 'datafile', 'csv' and 'excel'
    input_path: str
        Path to input data
    write_defaults: bool, default: False
        Expand default values to pad dataframes
    glpk_model : str
        Path to ``*.glp`` model file

    Returns
    -------
    Tuple[dict[str, pd.DataFrame], dict[str, float]]
        Dictionary of parameter and set data and dictionary of default values
    """
    user_config = _get_user_config(config)
    input_strategy = _get_read_strategy(user_config, input_format)
    result_strategy = _get_read_result_strategy(
        user_config, from_format, glpk_model, write_defaults
    )

    if input_strategy:
        input_data, _ = input_strategy.read(input_path)
    else:
        input_data = {}

    if result_strategy:
        results, default_values = result_strategy.read(from_path, input_data=input_data)
        return results, default_values
    else:
        msg = "Conversion from {} is not yet implemented".format(from_format)
        raise NotImplementedError(msg)


def convert_results(
    config: str,
    from_format: str,
    to_format: str,
    from_path: str,
    to_path: str,
    input_format: str,
    input_path: str,
    write_defaults: bool = False,
    glpk_model: Optional[str] = None,
) -> bool:
    """Post-process results from a CBC, CPLEX, Gurobi, HiHGS, or GLPK solution file into CSV format

    Arguments
    ---------
    config : str
        Path to config file
    from_format : str
        Available options are 'cbc', 'cplex', 'highs'. 'glpk', and 'gurobi'
    to_format : str
        Available options are 'csv', 'excel'
    from_path : str
        Path to solution file
    to_path : str
        Path to destination folder
    input_format: str
        Format of input data. Available options are 'datafile', 'csv' and 'excel'
    input_path: str
        Path to input data
    write_defaults: bool, default: False
        Expand default values to pad dataframes
    glpk_model : str
        Path to ``*.glp`` model file

    Returns
    -------
    bool
        True if conversion was successful, False otherwise

    """
    msg = "Conversion from {} to {} is not yet implemented".format(
        from_format, to_format
    )

    user_config = _get_user_config(config)

    # set read strategy

    read_strategy = _get_read_result_strategy(
        user_config, from_format, glpk_model, write_defaults
    )

    # set write strategy

    if to_format == "csv":
        write_strategy: WriteStrategy = WriteCsv(user_config=user_config)
    elif to_format == "excel":
        write_strategy = WriteExcel(user_config=user_config)
    else:
        raise NotImplementedError(msg)

    # read in input file
    input_data, _ = read(config, input_format, input_path)

    if read_strategy and write_strategy:
        context = Context(read_strategy, write_strategy)
        context.convert(from_path, to_path, input_data=input_data)
    else:
        raise NotImplementedError(msg)

    return True


def _get_read_result_strategy(
    user_config, from_format, glpk_model=None, write_defaults=False
) -> Union[ReadResults, None]:
    """Get ``ReadResults`` for gurobi, cbc, cplex, highs and glpk formats

    Arguments
    ---------
    config : dict
        User configuration describing parameters and sets
    from_format : str
        Available options are 'cbc', 'gurobi', 'cplex', 'highs', and 'glpk'
    write_defaults: bool, default: False
        Write default values to output format
    glpk_model : str
        Path to ``*.glp`` model file

    Returns
    -------
    ReadStrategy or None
        A ReadStrategy object. Returns None if from_format is not recognised

    """

    if from_format == "cbc":
        read_strategy: ReadResults = ReadCbc(
            user_config=user_config, write_defaults=write_defaults
        )
    elif from_format == "gurobi":
        read_strategy = ReadGurobi(
            user_config=user_config, write_defaults=write_defaults
        )
    elif from_format == "cplex":
        read_strategy = ReadCplex(
            user_config=user_config, write_defaults=write_defaults
        )
    elif from_format == "highs":
        read_strategy = ReadHighs(
            user_config=user_config, write_defaults=write_defaults
        )
    elif from_format == "glpk":
        if not glpk_model:
            raise OtooleError(resource="Read GLPK", message="Provide glpk model file")
        read_strategy = ReadGlpk(
            user_config=user_config,
            glpk_model=glpk_model,
            write_defaults=write_defaults,
        )
    else:
        return None

    return read_strategy


def _get_user_config(config) -> dict:
    """Read in the configuration file

    Arguments
    ---------
    config : str
        Path to config file

    Returns
    -------
    dict
        A dictionary containing the user configuration
    """
    if config:
        _, ending = os.path.splitext(config)
        with open(config, "r") as config_file:
            user_config = _read_file(config_file, ending)
        logger.info("Reading config from {}".format(config))
        logger.info("Validating config from {}".format(config))
        validate_config(user_config)
    return user_config


def _get_read_strategy(
    user_config, from_format, keep_whitespace=False, write_defaults=False
) -> ReadStrategy:
    """Get ``ReadStrategy`` for csv/datafile/excel format

    Arguments
    ---------
    config : dict
        User configuration describing parameters and sets
    from_format : str
        Available options are 'datafile', 'datapackage', 'csv' and 'excel'
    keep_whitespace: bool, default: False
        Keep whitespace in CSVs
    write_defaults: bool, default: False
        Expand default values to pad dataframes

    Returns
    -------
    ReadStrategy or None
        A ReadStrategy object. Returns None if from_format is not recognised

    """
    keep_whitespace = True if keep_whitespace else False

    if from_format == "datafile":
        read_strategy: ReadStrategy = ReadDatafile(
            user_config=user_config, write_defaults=write_defaults
        )
    elif from_format == "datapackage":
        logger.warning(
            "Reading from datapackage is deprecated, trying to read from CSVs"
        )
        logger.info("Successfully read folder of CSVs")
        read_strategy = ReadCsv(
            user_config=user_config,
            keep_whitespace=keep_whitespace,
            write_defaults=write_defaults,
        )  # typing: ReadStrategy
    elif from_format == "csv":
        read_strategy = ReadCsv(
            user_config=user_config,
            keep_whitespace=keep_whitespace,
            write_defaults=write_defaults,
        )  # typing: ReadStrategy
    elif from_format == "excel":
        read_strategy = ReadExcel(
            user_config=user_config,
            keep_whitespace=keep_whitespace,
            write_defaults=write_defaults,
        )  # typing: ReadStrategy
    else:
        msg = f"Conversion from {from_format} is not supported"
        raise NotImplementedError(msg)

    return read_strategy


def _get_write_strategy(user_config, to_format) -> WriteStrategy:
    """Get ``WriteStrategy`` for csv/datafile/excel format

    Arguments
    ---------
    user_config : dict
        User configuration describing parameters and sets
    to_format : str
        Available options are 'datafile', 'datapackage', 'csv' and 'excel'

    Returns
    -------
    WriteStrategy or None
        A ReadStrategy object. Returns None if to_format is not recognised

    """

    if to_format == "datapackage":
        write_strategy: WriteStrategy = WriteCsv(user_config=user_config)
    elif to_format == "excel":
        write_strategy = WriteExcel(user_config=user_config)
    elif to_format == "datafile":
        write_strategy = WriteDatafile(user_config=user_config)
    elif to_format == "csv":
        write_strategy = WriteCsv(user_config=user_config)
    else:
        msg = f"Conversion to {to_format} is not supported"
        raise NotImplementedError(msg)

    return write_strategy


def convert(
    config,
    from_format,
    to_format,
    from_path,
    to_path,
    write_defaults=False,
    keep_whitespace=False,
) -> bool:
    """Convert OSeMOSYS data from/to datafile, csv and Excel formats

    Arguments
    ---------
    config : str
        Path to config file
    from_format : str
        Available options are 'datafile', 'datapackage', 'csv' and 'excel'
    to_format : str
        Available options are 'datafile', 'datapackage', 'csv' and 'excel'
    from_path : str
        Path to destination file (if datafile or excel) or folder (csv or datapackage)
    write_defaults: bool, default: False
        Expand default values to pad dataframes
    keep_whitespace: bool, default: False
        Keep whitespace in CSVs

    Returns
    -------
    bool
        True if conversion was successful, False otherwise
    """

    user_config = _get_user_config(config)
    read_strategy = _get_read_strategy(
        user_config,
        from_format,
        keep_whitespace=keep_whitespace,
        write_defaults=write_defaults,
    )

    write_strategy = _get_write_strategy(user_config, to_format)

    if from_format == "datapackage":
        logger.warning(
            "Reading from and writing to datapackage is deprecated, writing to CSVs"
        )
        from_path = read_deprecated_datapackage(from_path)
        to_path = os.path.join(os.path.dirname(to_path), "data")

    context = Context(read_strategy, write_strategy)
    context.convert(from_path, to_path)

    return True


def read(
    config: str,
    from_format: str,
    from_path: str,
    keep_whitespace: bool = False,
    write_defaults: bool = False,
) -> Tuple[Dict[str, pd.DataFrame], Dict[str, float]]:
    """Read OSeMOSYS data from datafile, csv or Excel formats

    Arguments
    ---------
    config : str
        Path to config file
    from_format : str
        Available options are 'datafile', 'csv', 'excel' and 'datapackage' [deprecated]
    from_path : str
        Path to source file (if datafile or excel) or folder (csv)
    keep_whitespace: bool, default: False
        Keep whitespace in source files
    write_defaults: bool, default: False
        Expand default values to pad dataframes

    Returns
    -------
    Tuple[dict[str, pd.DataFrame], dict[str, float]]
        Dictionary of parameter and set data and dictionary of default values
    """
    user_config = _get_user_config(config)
    read_strategy = _get_read_strategy(
        user_config,
        from_format,
        keep_whitespace=keep_whitespace,
        write_defaults=write_defaults,
    )

    if from_format == "datapackage":
        from_path = read_deprecated_datapackage(from_path)

    return read_strategy.read(from_path)


def write(
    config: str,
    to_format: str,
    to_path: str,
    inputs,
    default_values: Optional[Dict[str, float]] = None,
) -> bool:
    """Write OSeMOSYS data to datafile, csv or Excel formats

    Arguments
    ---------
    config : str
        Path to config file
    to_format : str
        Available options are 'datafile', 'csv', 'excel' and 'datapackage' [deprecated],
    to_path : str
        Path to destination file (if datafile or excel) or folder (csv))
    inputs : dict[str, pd.DataFrame]
        Dictionary of pandas data frames to write
    default_values: dict[str, float], default: None
        Dictionary of default values to write to datafile

    """
    user_config = _get_user_config(config)
    write_strategy = _get_write_strategy(user_config, to_format)
    if default_values is None:
        write_strategy.write(inputs, to_path, {})
    else:
        write_strategy.write(inputs, to_path, default_values)

    return True
